"""API for recorded game playback."""

import asyncio
import logging
import os
import shutil
import struct
import time
from enum import Enum

import aiohttp
import flatbuffers
import tqdm

from AOC import ConfigMessage
from AOC import GameMessage

from mgz import fast


LOGGER = logging.getLogger(__name__)
WS_URL = 'ws://{}'
MAX_ATTEMPTS = 5
TOTAL_TIMEOUT = 60 * 4
IO_TIMEOUT = 10


class Source(Enum):
    """Message source."""
    MGZ = 0
    MEMORY = 1


def make_config(interval=1000, cycles=1):
    """Make a configuration flatbuffer."""
    builder = flatbuffers.Builder(64)
    ConfigMessage.ConfigMessageStart(builder)
    ConfigMessage.ConfigMessageAddMessageIntervalMs(builder, interval)
    ConfigMessage.ConfigMessageAddUpdateCycles(builder, cycles)
    builder.Finish(ConfigMessage.ConfigMessageEnd(builder))
    return bytes(builder.Output())


def read_message(buf):
    """Read a message from flatbuffer as a dict."""
    return GameMessage.GameMessage.GetRootAsGameMessage(buf, 0)


async def progress_bar(generator, duration):
    """Apply a progress bar - visualize progress in console."""
    config = {
        'total': duration,
        'unit': ' ticks',
        'miniters': int(duration/10000),
        'mininterval': 1
    }
    with tqdm.tqdm(**config) as bar_:
        last = 0
        async for tick, source, data in generator:
            increment = tick - last
            if increment > 0:
                bar_.update(increment)
                last = tick
            yield tick, source, data


class Client():
    """Represents a client instance."""

    def __init__(self, playback, rec_path):
        """Initialize."""
        parts = playback.split(':')
        self.socket = ':'.join(parts[0:-1])
        self.dropbox_path = os.path.abspath(os.path.expanduser(parts[-1]))
        self.rec_path = os.path.abspath(os.path.expanduser(rec_path))
        self._handle = open(self.rec_path, 'rb')
        header_len, _ = struct.unpack('<II', self._handle.read(8))
        self._handle.read(header_len - 8)

    # pylint: disable=attribute-defined-outside-init, too-many-arguments
    @classmethod
    async def create(
            cls, playback, rec_path, start_time, duration,
            interval=1000, cycles=10000, total_timeout=TOTAL_TIMEOUT, io_timeout=IO_TIMEOUT
    ):
        """Async factory."""
        self = Client(playback, rec_path)
        self.start_time = start_time
        self.duration = duration
        self.total_timeout = total_timeout
        self.io_timeout = io_timeout
        self.session, self.state_ws = await self.start_instance(interval, cycles)
        LOGGER.info("launched instance for %s", rec_path)
        return self

    async def start_instance(self, interval, cycles):
        """Start headless instance and connect."""
        shutil.copyfile(
            self.rec_path,
            os.path.join(self.dropbox_path, os.path.basename(self.rec_path))
        )
        # wait for any existing AOC to be killed so we don't connect to wrong WS
        await asyncio.sleep(2)
        attempts = 0
        url = WS_URL.format(self.socket)
        LOGGER.info("trying to connect @ %s", url)
        session = aiohttp.ClientSession()
        while attempts < MAX_ATTEMPTS:
            try:
                websocket = await asyncio.wait_for(session.ws_connect(url), timeout=self.io_timeout)
                try:
                    await asyncio.wait_for(websocket.send_bytes(make_config(interval, cycles)), timeout=self.io_timeout)
                except asyncio.TimeoutError:
                    LOGGER.error("failed to send configuration")
                    break
                LOGGER.info("sent configuration to websocket")
                return session, websocket
            except (aiohttp.client_exceptions.ClientError, asyncio.TimeoutError):
                LOGGER.info("trying again to connect ...")
                await asyncio.sleep(1)
                attempts += 1
        LOGGER.error("failed to launch playback")
        self._handle.close()
        await session.close()
        raise RuntimeError("failed to launch playback")

    async def read_state(self):
        """Read memory state messages."""
        async for message in self.state_ws:
            yield read_message(message.data)

    async def sync(self):
        """Synchronize state and body messages."""
        mgz_time = self.start_time
        ws_time = 0
        mgz_done = False
        ws_done = False
        LOGGER.info("starting synchronization")
        start = time.time()
        keep_reading = True
        while not mgz_done or not ws_done:
            if self.total_timeout and time.time() - start > self.total_timeout:
                LOGGER.warning("playback time exceeded timeout (%d%% completed)", int(mgz_time/self.duration * 100))
                break
            try:
                while not mgz_done and (mgz_time <= ws_time or ws_done):
                    op_type, payload = fast.operation(self._handle)
                    if op_type == fast.Operation.SYNC:
                        mgz_time += payload[0]
                    elif op_type == fast.Operation.CHAT and payload:
                        yield mgz_time, Source.MGZ, (op_type, payload.strip(b'\x00'))
                    elif op_type == fast.Operation.ACTION:
                        yield mgz_time, Source.MGZ, (op_type, payload)
            except EOFError:
                LOGGER.info("MGZ parsing stream finished")
                mgz_done = True
            while not ws_done and (ws_time <= mgz_time or mgz_done):
                try:
                    data = await asyncio.wait_for(self.read_state().__anext__(), timeout=self.io_timeout)
                except (asyncio.TimeoutError, asyncio.streams.IncompleteReadError):
                    LOGGER.warning("state reader timeout")
                    keep_reading = False
                    break
                ws_time = data.WorldTime()
                if data.GameFinished():
                    LOGGER.info("state reader stream finished")
                    ws_done = True
                yield ws_time, Source.MEMORY, data
            if not keep_reading:
                break
        if ws_done and mgz_done:
            LOGGER.info("synchronization finished in %.2f seconds", time.time() - start)
        else:
            raise RuntimeError("synchronization timeout encountered")
        await self.session.close()
        self._handle.close()



"""Determine dataset."""

import mgz
from mgz.util import Version


def get_dataset_data(header):
    """Get dataset."""
    if header.version == Version.DE:
        return {
            'id': 100,
            'name': 'Definitive Edition',
            'version': header.game_version[4:]
        }
    sample = header.initial.players[0].attributes.player_stats
    if 'mod' in sample and sample.mod['id'] == 0 and sample.mod['version'] == '2':
        raise ValueError("invalid mod version")
    if 'mod' in sample and sample.mod['id'] > 0:
        return sample.mod
    if 'trickle_food' in sample and sample.trickle_food:
        return {
            'id': 1,
            'name': mgz.const.MODS.get(1),
            'version': '<5.7.2'
        }
    if header.version == Version.AOK:
        return {
            'id': 200,
            'name': 'Age of Kings',
            'version': '2.0a'
        }
    return {
        'id': 0,
        'name': 'The Conquerors',
        'version': '1.0c'
    }



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: AOC

import flatbuffers

class GameMessage(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsGameMessage(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = GameMessage()
        x.Init(buf, n + offset)
        return x

    # GameMessage
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # GameMessage
    def StateReaderVersion(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # GameMessage
    def GameIndex(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # GameMessage
    def WorldTime(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # GameMessage
    def GameFinished(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # GameMessage
    def MarketCoefficients(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = o + self._tab.Pos
            from .MarketCoefficients import MarketCoefficients
            obj = MarketCoefficients()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # GameMessage
    def Players(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .Player import Player
            obj = Player()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # GameMessage
    def PlayersLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # GameMessage
    def Objects(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 68
            from .Object import Object
            obj = Object()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # GameMessage
    def ObjectsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # GameMessage
    def StateChanges(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 12
            from .StateChange import StateChange
            obj = StateChange()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # GameMessage
    def StateChangesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

def GameMessageStart(builder): builder.StartObject(8)
def GameMessageAddStateReaderVersion(builder, stateReaderVersion): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(stateReaderVersion), 0)
def GameMessageAddGameIndex(builder, gameIndex): builder.PrependInt32Slot(1, gameIndex, 0)
def GameMessageAddWorldTime(builder, worldTime): builder.PrependUint32Slot(2, worldTime, 0)
def GameMessageAddGameFinished(builder, gameFinished): builder.PrependBoolSlot(3, gameFinished, 0)
def GameMessageAddMarketCoefficients(builder, marketCoefficients): builder.PrependStructSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(marketCoefficients), 0)
def GameMessageAddPlayers(builder, players): builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(players), 0)
def GameMessageStartPlayersVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def GameMessageAddObjects(builder, objects): builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(objects), 0)
def GameMessageStartObjectsVector(builder, numElems): return builder.StartVector(68, numElems, 4)
def GameMessageAddStateChanges(builder, stateChanges): builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(stateChanges), 0)
def GameMessageStartStateChangesVector(builder, numElems): return builder.StartVector(12, numElems, 4)
def GameMessageEnd(builder): return builder.EndObject()



"""Determine teams."""


def get_teams_data(header):
    """Get teams."""
    teams = []
    for j, player in enumerate(header.initial.players):
        added = False
        for i in range(0, len(header.initial.players)):
            if player.attributes.my_diplomacy[i] == 'ally':
                inner_team = False
                outer_team = False
                new_team = True
                for k, members in enumerate(teams):
                    if j in members or i in members:
                        new_team = False
                    if j in members and i not in members:
                        inner_team = k
                        break
                    if j not in members and i in members:
                        outer_team = k
                        break
                if new_team:
                    teams.append([i, j])
                if inner_team is not False:
                    teams[inner_team].append(i)
                if outer_team is not False:
                    teams[outer_team].append(j)
                added = True
        if not added and j != 0:
            teams.append([j])
    return teams



"""Body.

An mgz body is a stream of Operations

An Operation can be:
 - Action: Player input that materially affects the game
 - Message: Either a start-of-game indicator, or chat
 - Synchronization: Time increment and view coordinates of recording player
 - Embedded: A variety of embedded structures
"""

from construct import (Struct, Byte, Switch, Embedded, Padding,
                       Int32ul, Peek, Tell, Float32l, String, If, Array, Bytes,
                       GreedyBytes, Computed, IfThenElse, Int16ul, Int64ul)
from mgz.enums import ActionEnum, OperationEnum, MessageEnum
from mgz.body import actions, embedded
from mgz.util import BoolAdapter


# pylint: disable=invalid-name, bad-continuation


# Action.
action_data = "action"/Struct(
    Peek("type_int"/Byte),
    ActionEnum("type"/Byte),
    Embedded("action"/Switch(lambda ctx: ctx.type, {
        "interact": actions.interact,
        "move": actions.move,
        "stop": actions.stop,
        "create": actions.create,
        "stance": actions.stance,
        "guard": actions.guard,
        "follow": actions.follow,
        "formation": actions.formation,
        "waypoint": actions.waypoint,
        "give_attribute": actions.give_attribute,
        "add_attribute": actions.add_attribute,
        "ai_waypoint": actions.ai_waypoint,
        "ai_interact": actions.ai_interact,
        "ai_move": actions.ai_move,
        "ai_queue": actions.ai_queue,
        "save": actions.save,
        "chapter": actions.chapter,
        "ai_command": actions.ai_command,
        "spec": actions.spec,
        "build": actions.build,
        "game": actions.game,
        "patrol": actions.patrol,
        "wall": actions.wall,
        "delete": actions.delete,
        "attackground": actions.attackground,
        "repair": actions.repair,
        "release": actions.release,
        "togglegate": actions.togglegate,
        "flare": actions.flare,
        "order": actions.order,
        "droprelic": actions.droprelic,
        "gatherpoint": actions.gatherpoint,
        "townbell": actions.townbell,
        "resign": actions.resign,
        "tribute": actions.tribute,
        "queue": actions.queue,
        "multiqueue": actions.multiqueue,
        "research": actions.research,
        "sell": actions.sell,
        "buy": actions.buy,
        "backtowork": actions.backtowork,
        "de_queue": actions.de_queue,
        "postgame": actions.postgame,
        "de_attackmove": actions.de_attackmove,
        "de_autoscout": actions.de_autoscout
    }, default=Struct(
        "unk_action"/Computed(lambda ctx: ctx._.type),
        "bytes"/Bytes(lambda ctx: ctx._._.length - 1)
    ))),
    Padding(4)
)


# Action - length followed by data.
action = "action"/Struct(
    "length"/Int32ul,
    action_data
)


# Synchronization.
sync = "sync"/Struct(
    "time_increment"/Int32ul,
    Peek("next"/Int32ul),
    "checksum"/If(lambda ctx: not ctx.next, Struct(
        Padding(8),
        "sync"/Int32ul,
        "unknown"/Bytes(4),
        "sequence"/Int32ul,
        If(lambda ctx: ctx.sequence > 0, Bytes(332)),
        Bytes(8)
    ))
)


# View lock.
viewlock = "viewlock"/Struct(
    "view"/Struct(
        "x"/Float32l,
        "y"/Float32l
    ),
    "player_id"/Int32ul
)


# Chat variation of Message.
chat = "chat"/Struct(
    Padding(4),
    "length"/Int32ul,
    "text"/Bytes(lambda ctx: ctx.length - 0),
    Padding(0)
)


# Start
start = "start"/Struct(
    "checksum_interval"/Int32ul,
    BoolAdapter("multiplayer"/Int32ul),
    "rec_owner"/Int32ul,
    BoolAdapter("reveal_map"/Int32ul),
    "use_sequence_numbers"/Int32ul,
)


# AOK start.
aok_start = "aok_start"/Struct(
    Embedded(start),
    "empires2_exe_size"/Int64ul,
    Padding(4)
)


# AOC start.
aoc_start = "aoc_start"/Struct(
    Embedded(start),
    "number_of_chapters"/Int32ul
)


# DE start.
de_start = "de_start"/Struct(
    Padding(4),
    Embedded(aoc_start)
)


# Message.
message = "message"/Struct(
    MessageEnum("subtype"/Peek(Int32ul)),
    "data"/Switch(lambda ctx: ctx.subtype, {
        "start": aoc_start,
        "chat": chat
    })
)


# Operation.
operation = "operation"/Struct(
    Peek(OperationEnum("type"/Int32ul)),
    "start"/Tell,
    "op"/Int32ul,
    Embedded("data"/Switch(lambda ctx: ctx.type, {
        "action": action,
        "sync": sync,
        "viewlock": viewlock,
        "message": message,
        "de_start": de_start,
        "aok_start": aok_start,
        "embedded": embedded.embedded
    })),
    "end"/Tell
)



"""MGZ parsing utilities."""

import logging
import struct
import zlib
from enum import Enum
from io import BytesIO

import construct.core
from construct import Adapter, Construct, Subconstruct, Tunnel

from mgz import const


# pylint: disable=abstract-method,protected-access

LOGGER = logging.getLogger(__name__)
SEARCH_MAX_BYTES = 3000
POSTGAME_LENGTH = 2096
LOOKAHEAD = 9


class Version(Enum):
    """Version enumeration.

    Using consts from https://github.com/goto-bus-stop/recanalyst/blob/master/src/Model/Version.php
    for consistency.
    """
    AOK = 1
    AOC = 5
    AOC10C = 8
    USERPATCH12 = 12
    USERPATCH13 = 13
    USERPATCH14 = 11
    USERPATCH15 = 20
    DE = 21
    USERPATCH14RC2 = 22


class MgzPrefixed(Subconstruct):
    """Like `Prefixed`, but accepting arbitrary length."""

    __slots__ = ["name", "length", "subcon"]

    def __init__(self, length, subcon):
        """Initialize."""
        super(MgzPrefixed, self).__init__(subcon)
        self.length = length

    def _parse(self, stream, context, path):
        """Parse tunnel."""
        length = self.length(context)
        new_stream = BytesIO(construct.core._read_stream(stream, length))
        return self.subcon._parse(new_stream, context, path)


class ZlibCompressed(Tunnel):
    """Like Compressed, but only does header-less zlib."""

    __slots__ = []

    def _decode(self, data, context):
        """Decode zlib without header bytes."""
        return zlib.decompress(data, wbits=-15)


def get_version(game_version, save_version):
    """Get version based on version fields."""
    if game_version == 'VER 9.3':
        return Version.AOK
    if game_version == 'VER 9.4':
        if save_version >= 12.97:
            return Version.DE
        return Version.AOC
    if game_version == 'VER 9.8':
        return Version.USERPATCH12
    if game_version == 'VER 9.9':
        return Version.USERPATCH13
    if game_version == 'VER 9.A':
        return Version.USERPATCH14RC2
    if game_version in ['VER 9.B', 'VER 9.C', 'VER 9.D']:
        return Version.USERPATCH14
    if game_version in ['VER 9.E', 'VER 9.F']:
        return Version.USERPATCH15
    raise ValueError('unsupported version: {}, {}'.format(game_version, save_version))


def check_flags(peek):
    """Check byte sequence for only flag bytes."""
    for i in peek:
        if i not in [0, 1]:
            return False
    return True


def convert_to_timestamp(time):
    """Convert int to timestamp string."""
    if time == -1:
        return None
    time = int(time*1000)
    hour = time//1000//3600
    minute = (time//1000//60) % 60
    second = (time//1000) % 60
    return str(hour).zfill(2)+":"+str(minute).zfill(2)+":"+str(second).zfill(2)


class TimeSecAdapter(Adapter):
    """Conversion to readable time."""

    def _decode(self, obj, context):
        """Decode timestamp to string."""
        return convert_to_timestamp(obj)


class BoolAdapter(Adapter):
    """Bools with potential padding."""

    def _decode(self, obj, context):
        """Decode bool."""
        return obj == 1


class VersionAdapter(Adapter):
    """Round save version."""

    def _decode(self, obj, context):
        """Decode by rounding float."""
        return round(obj, 2)


class ModVersionAdapter(Adapter):
    """Parse mod version."""

    def _decode(self, obj, context):
        """Decode mod."""
        number = int(obj)
        mod_id = int(number / 1000)
        mod_version = '.'.join(list(str(number % 1000)))
        return {
            'id': mod_id,
            'name': const.MODS.get(mod_id, 'Unknown Mod'),
            'version': mod_version
        }


class Find(Construct):
    """Find bytes, and read past them."""

    __slots__ = ["find", "max_length"]

    def __init__(self, find, max_length):
        """Initiallize."""
        Construct.__init__(self)
        self.find = find
        self.max_length = max_length

    def _parse(self, stream, context, path):
        """Parse stream to find a given byte string."""
        start = stream.tell()
        read_bytes = ""
        if self.max_length:
            read_bytes = stream.read(self.max_length)
        else:
            read_bytes = stream.read()
        skip = read_bytes.find(self.find) + len(self.find)
        stream.seek(start + skip)
        return skip


class RepeatUpTo(Subconstruct):
    """Like RepeatUntil, but doesn't include the last element in the return value."""

    __slots__ = ["find"]

    def __init__(self, find, subcon):
        """Initialize."""
        Subconstruct.__init__(self, subcon)
        self.find = find

    def _parse(self, stream, context, path):
        """Parse until a given byte string is found."""
        objs = []
        while True:
            start = stream.tell()
            test = stream.read(len(self.find))
            stream.seek(start)
            if test == self.find:
                break
            subobj = self.subcon._parse(stream, context, path)
            objs.append(subobj)
        return objs


class GotoObjectsEnd(Construct):
    """Find the end of a player's objects list.

    Necessary since we can't parse objects from a resume game (yet).
    """

    # pylint: disable=chained-comparison, too-many-locals
    def _parse(self, stream, context, path):
        """Parse until the end of objects data."""
        num_players = context._._.replay.num_players
        marker_num = context.attributes.num_header_data
        save_version = context._._.save_version
        start = stream.tell()
        # Have to read everything to be able to use find()
        read_bytes = stream.read()
        # Try to find the first marker, a portion of the next player structure
        # The byte that changes is the number of player stats fields
        marker = read_bytes.find(b'\x16' + struct.pack('<I', int(marker_num)) + b'\x21')
        # If it exists, we're not on the last player yet
        if marker > 0:
            # Backtrack through the player name
            count = 0
            while struct.unpack("<H", read_bytes[marker-2:marker])[0] != count:
                marker -= 1
                count += 1
            # Backtrack through the rest of the next player structure
            backtrack = 43 + num_players
        # Otherwise, this is the last player
        else:
            # Search for the scenario header
            # TODO: make this section more reliable
            marker_aok = read_bytes.find(b"\x9a\x99\x99\x3f")
            marker_up = read_bytes.find(b"\xf6\x28\x9c\x3f")
            if save_version >= 13.07:
                marker_de = read_bytes.find(b"\x29\x5c\xaf\x3f")
            else:
                marker_de = read_bytes.find(b"\x7b\x14\xae\x3f")
            if marker_up > 0 and marker_de < 0: # aok marker can appear in up
                marker = marker_up
            elif marker_de > 0 and marker_up < 0 and marker_aok < 0:
                marker = marker_de
            elif marker_aok > 0 and marker_up < 0 and marker_de < 0:
                marker = marker_aok
            # Backtrack through the achievements and initial structure footer
            backtrack = ((1817 * (num_players - 1)) + 4 + 19)
        # Seek to the position we found
        end = start + marker - backtrack - 2
        stream.seek(end)
        return end


def find_postgame(data, size):
    """Find postgame and grab duration.

    We can find postgame location by scanning the last few
    thousand bytes of the rec and looking for a pattern as
    follows:

    [action op]    [action length]    [action type]
    01 00 00 00    30 08 00 00        ff

    The last occurance of this pattern signals the start of
    the postgame structure. Note that the postgame action length
    is always constant, unlike other actions.
    """
    pos = None
    for i in range(size - SEARCH_MAX_BYTES, size - LOOKAHEAD):
        op_type, length, action_type = struct.unpack('<IIB', data[i:i + LOOKAHEAD])
        if op_type == 0x01 and length == POSTGAME_LENGTH and action_type == 0xFF:
            LOGGER.debug("found postgame candidate @ %d with length %d", pos, length)
            return i + LOOKAHEAD, length
    return None, None



"""Lobby."""

from construct import Array, Byte, Bytes, Flag, Int32ul, Padding, Peek, Struct, If, Computed, Embedded

from mgz.enums import GameTypeEnum, RevealMapEnum
from mgz.util import Version

# pylint: disable=invalid-name, bad-continuation


# Player inputs in the lobby, and several host settings.
lobby = "lobby"/Struct(
    Array(8, "teams"/Byte), # team number selected by each player
    If(lambda ctx: ctx._.version != Version.DE,
        Padding(1),
    ),
    Peek("reveal_map_id"/Int32ul),
    RevealMapEnum("reveal_map"/Int32ul),
    Padding(4),
    "map_size"/Int32ul,
    "population_limit_encoded"/Int32ul,
    "population_limit"/Computed(lambda ctx: ctx.population_limit_encoded * (25 if ctx._.version in [Version.USERPATCH14, Version.USERPATCH15] else 1)),
    Embedded(If(lambda ctx: ctx._.version != Version.AOK,
        Struct(
            Peek("game_type_id"/Byte),
            GameTypeEnum("game_type"/Byte),
            "lock_teams"/Flag
        )
    )),
    If(lambda ctx: ctx._.version == Version.DE,
        Padding(5)
    ),
    Embedded(If(lambda ctx: ctx._.version != Version.AOK,
        Struct(
            "num_chat"/Int32ul,
            Array(
                lambda ctx: ctx.num_chat, # pre-game chat messages
                "messages"/Struct(
                    "message_length"/Int32ul,
                    "message"/Bytes(lambda ctx: ctx.message_length - (1 if ctx.message_length > 0 else 0)),
                    If(lambda ctx: ctx.message_length > 0, Padding(1))
                )
            )
        )
    )),
    If(lambda ctx: ctx._.version == Version.DE,
        Padding(10)
    )
)



"""Definitive Edition structure."""

from construct import (
    Struct, Int32ul, Float32l, Array, Padding, Flag,
    Byte, Int16ul, Bytes, Int32sl, Peek, Const, RepeatUntil
)

from mgz.enums import VictoryEnum, ResourceLevelEnum, AgeEnum

# pylint: disable=invalid-name, bad-continuation

de = "de"/Struct(
    "options"/Int32sl,
    "unk0"/Int32ul,
    "unk1"/Int32ul,
    "dlc_count"/Int32ul,
    "dlc_ids"/Array(lambda ctx: ctx.dlc_count, Int32ul),
    Padding(4),
    "difficulty"/Int32ul,
    "map_size"/Int32ul,
    "map_id"/Int32ul,
    "reveal_map"/Int32ul,
    Peek("victory_type_id"/Int32ul),
    VictoryEnum("victory_type"/Int32ul),
    Peek("starting_resources_id"/Int32ul),
    ResourceLevelEnum("starting_resources"/Int32ul),
    Peek("starting_age_id"/Int32ul),
    AgeEnum("starting_age"/Int32ul),
    Peek("ending_age_id"/Int32ul),
    AgeEnum("ending_age"/Int32ul),
    "game_type"/Int32ul,
    Padding(8),
    "speed"/Float32l,
    "treaty_length"/Int32ul,
    "population_limit"/Int32ul,
    "num_players"/Int32ul,
    "unknown2"/Int32ul,
    "unknown3"/Int32ul,
    Padding(4),
    "trade_enabled"/Flag,
    "team_bonus_disabled"/Flag,
    "random_positions"/Flag,
    "all_techs"/Flag,
    "num_starting_units"/Byte,
    "lock_teams"/Flag,
    "lock_speed"/Flag,
    "multiplayer"/Flag,
    "cheats"/Flag,
    "record_game"/Flag,
    "animals_enabled"/Flag,
    "predators_enabled"/Flag,
    Padding(6),
    "players"/Array(8, Struct(
        "dat_crc"/Bytes(4),
        "mp_version"/Byte,
        "color_id"/Int32sl,
        Bytes(12),
        "civ_id"/Byte,
        Bytes(14),
        "name_length"/Int16ul,
        "name"/Bytes(lambda ctx: ctx.name_length),
        Padding(4),
        "profile_id"/Int32ul,
        Bytes(17)
    )),
    Padding(1),
    "fog_of_war"/Flag,
    "cheat_notifications"/Flag,
    "colored_chat"/Flag,
    Padding(4),
    "ranked"/Flag,
    "allow_specs"/Flag,
    Padding(19),
    "strings"/Array(23,
        Struct(
            Const(b"\x60\x0A"),
            "len"/Int16ul,
            "string"/Bytes(lambda ctx: ctx.len),
            RepeatUntil(lambda x, lst, ctx: lst[-1] not in [3, 21, 42], Int32ul)
        )
    ),
    Padding(244),
    "guid"/Bytes(16),
    Padding(2),
    "lobby_name_length"/Int16ul,
    "lobby_name"/Bytes(lambda ctx: ctx.lobby_name_length),
    Padding(44)
)



"""Determine player data."""


def ach(structure, fields):
    """Get field from achievements structure."""
    field = fields.pop(0)
    if structure:
        if hasattr(structure, field):
            structure = getattr(structure, field)
            if not fields:
                return structure
            return ach(structure, fields)
    return None


def get_achievements(postgame, name):
    """Get achievements for a player.

    Must match on name, not index, since order is not always the same.
    """
    if not postgame:
        return None
    for achievements in postgame.achievements:
        # achievements player name can be shorter
        if name.startswith(achievements.player_name.replace(b'\x00', b'')):
            return achievements
    return None


def guess_winner(teams, resigned, i):
    """Guess if a player won.

    Find what team the player was on. If anyone
    on their team resigned, assume the player lost.
    """
    for team in teams:
        if i not in team:
            continue
        for j in team:
            if j in resigned:
                return False
    return True


def get_players_data(header, postgame, teams, resigned, cheaters, profile_ids, ratings, encoding): # pylint: disable=too-many-arguments, too-many-locals
    """Get basic player info."""
    out = []
    for i, player in enumerate(header.initial.players[1:]):
        achievements = get_achievements(postgame, player.attributes.player_name)
        if achievements:
            winner = achievements.victory
        else:
            winner = guess_winner(teams, resigned, i + 1)
        feudal_time = ach(achievements, ['technology', 'feudal_time_int'])
        castle_time = ach(achievements, ['technology', 'castle_time_int'])
        imperial_time = ach(achievements, ['technology', 'imperial_time_int'])
        name = player.attributes.player_name.decode(encoding)
        out.append({
            'name': name,
            'civilization': player.attributes.civilization,
            'human': header.scenario.game_settings.player_info[i + 1].type == 'human',
            'number': i + 1,
            'color_id': player.attributes.player_color,
            'winner': winner,
            'mvp': ach(achievements, ['mvp']),
            'score': ach(achievements, ['total_score']),
            'position': (player.attributes.camera_x, player.attributes.camera_y),
            'rate_snapshot': ratings.get(name),
            'user_id': profile_ids.get(player.attributes.player_color),
            'cheater': (i + 1) in cheaters,
            'achievements': {
                'military': {
                    'score': ach(achievements, ['military', 'score']),
                    'units_killed': ach(achievements, ['military', 'units_killed']),
                    'hit_points_killed': ach(achievements, ['military', 'hit_points_killed']),
                    'units_lost': ach(achievements, ['military', 'units_lost']),
                    'buildings_razed': ach(achievements, ['military', 'buildings_razed']),
                    'hit_points_razed': ach(achievements, ['military', 'hit_points_razed']),
                    'buildings_lost': ach(achievements, ['military', 'buildings_lost']),
                    'units_converted': ach(achievements, ['military', 'units_converted'])
                },
                'economy': {
                    'score': ach(achievements, ['economy', 'score']),
                    'food_collected': ach(achievements, ['economy', 'food_collected']),
                    'wood_collected': ach(achievements, ['economy', 'wood_collected']),
                    'stone_collected': ach(achievements, ['economy', 'stone_collected']),
                    'gold_collected': ach(achievements, ['economy', 'gold_collected']),
                    'tribute_sent': ach(achievements, ['economy', 'tribute_sent']),
                    'tribute_received': ach(achievements, ['economy', 'tribute_received']),
                    'trade_gold': ach(achievements, ['economy', 'trade_gold']),
                    'relic_gold': ach(achievements, ['economy', 'relic_gold'])
                },
                'technology': {
                    'score': ach(achievements, ['technology', 'score']),
                    'feudal_time': feudal_time if feudal_time and feudal_time > 0 else None,
                    'castle_time': castle_time if castle_time and castle_time > 0 else None,
                    'imperial_time': imperial_time if imperial_time and imperial_time > 0 else None,
                    'explored_percent': ach(achievements, ['technology', 'explored_percent']),
                    'research_count': ach(achievements, ['technology', 'research_count']),
                    'research_percent': ach(achievements, ['technology', 'research_percent'])
                },
                'society': {
                    'score': ach(achievements, ['society', 'score']),
                    'total_wonders': ach(achievements, ['society', 'total_wonders']),
                    'total_castles': ach(achievements, ['society', 'total_castles']),
                    'total_relics': ach(achievements, ['society', 'relics_captured']),
                    'villager_high': ach(achievements, ['society', 'villager_high'])
                }
            }
        })
    return out



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: AOC

import flatbuffers

class VictoryPoints(object):
    __slots__ = ['_tab']

    # VictoryPoints
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # VictoryPoints
    def RelicsCaptured(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # VictoryPoints
    def RelicGold(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(4))
    # VictoryPoints
    def TradeProfit(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(8))
    # VictoryPoints
    def TributeReceived(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(12))
    # VictoryPoints
    def TributeSent(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(16))
    # VictoryPoints
    def TotalFood(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(20))
    # VictoryPoints
    def TotalWood(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(24))
    # VictoryPoints
    def TotalGold(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(28))
    # VictoryPoints
    def TotalStone(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(32))
    # VictoryPoints
    def ValueSpentObjects(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(36))
    # VictoryPoints
    def ValueSpentResearch(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(40))
    # VictoryPoints
    def ValueLostUnits(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(44))
    # VictoryPoints
    def ValueLostBuildings(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(48))
    # VictoryPoints
    def ValueCurrentUnits(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(52))
    # VictoryPoints
    def ValueCurrentBuildings(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(56))
    # VictoryPoints
    def ValueObjectsDestroyed(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(60))

def CreateVictoryPoints(builder, relicsCaptured, relicGold, tradeProfit, tributeReceived, tributeSent, totalFood, totalWood, totalGold, totalStone, valueSpentObjects, valueSpentResearch, valueLostUnits, valueLostBuildings, valueCurrentUnits, valueCurrentBuildings, valueObjectsDestroyed):
    builder.Prep(4, 64)
    builder.PrependFloat32(valueObjectsDestroyed)
    builder.PrependFloat32(valueCurrentBuildings)
    builder.PrependFloat32(valueCurrentUnits)
    builder.PrependFloat32(valueLostBuildings)
    builder.PrependFloat32(valueLostUnits)
    builder.PrependFloat32(valueSpentResearch)
    builder.PrependFloat32(valueSpentObjects)
    builder.PrependFloat32(totalStone)
    builder.PrependFloat32(totalGold)
    builder.PrependFloat32(totalWood)
    builder.PrependFloat32(totalFood)
    builder.PrependFloat32(tributeSent)
    builder.PrependFloat32(tributeReceived)
    builder.PrependFloat32(tradeProfit)
    builder.PrependFloat32(relicGold)
    builder.PrependFloat32(relicsCaptured)
    return builder.Offset()



"""Determine diplomacy type."""


def get_diplomacy_data(header, teams):
    """Compute diplomacy."""
    player_num = 0
    computer_num = 0
    for player in header.scenario.game_settings.player_info:
        if player.type == 'human':
            player_num += 1
        elif player.type == 'computer':
            computer_num += 1
    total_num = player_num + computer_num

    diplomacy = {
        'FFA': (len(teams) == total_num) and total_num > 2,
        'TG':  len(teams) == 2 and total_num > 2,
        '1v1': total_num == 2,
    }

    diplomacy['type'] = 'Other'
    team_sizes = sorted([len(team) for team in teams])
    diplomacy['team_size'] = 'v'.join([str(size) for size in team_sizes])
    if diplomacy['FFA']:
        diplomacy['type'] = 'FFA'
        diplomacy['team_size'] = 'FFA'
    elif diplomacy['TG']:
        diplomacy['type'] = 'TG'
    elif diplomacy['1v1']:
        diplomacy['type'] = '1v1'
    return diplomacy



""""Extract data via playback."""

import logging
import time
from collections import defaultdict
from datetime import timedelta

from mgz import fast
from mgz.playback import Client, Source


LOGGER = logging.getLogger(__name__)
CLASS_UNIT = 70
CLASS_BUILDING = 80
TECH_STATE_AVAILABLE = 2
TECH_STATE_RESEARCHING = 3
TECH_STATE_DONE = 4


def has_diff(data, **kwargs):
    """Simple dict subset diff."""
    for key, value in kwargs.items():
        if data.get(key) != value:
            return True
    return False


def flatten_research(research):
    """Flatten research data."""
    flat = []
    for pid, techs in research.items():
        for tid, values in techs.items():
            flat.append(dict(values, player_number=pid, technology_id=tid))
    return flat


def build_timeseries_record(tick, player):
    """Build a timeseries record."""
    return {
        'timestamp': tick,
        'player_number': player.Id(),
        'population': player.Population(),
        'military': player.MilitaryPopulation(),
        'percent_explored': player.PercentMapExplored(),
        'headroom': int(player.Headroom()),
        'food': int(player.Food()),
        'wood': int(player.Wood()),
        'stone': int(player.Stone()),
        'gold': int(player.Gold()),
        'total_housed_time': int(player.CumulativeHousedTime()),
        'total_popcapped_time': int(player.CumulativePopCappedTime()),
        'relics_captured': int(player.VictoryPointsAndAttributes().VpRelicsCaptured()),
        'relic_gold': int(player.VictoryPointsAndAttributes().VpRelicGold()),
        'trade_profit': int(player.VictoryPointsAndAttributes().VpTradeProfit()),
        'tribute_received': int(player.VictoryPointsAndAttributes().VpTributeReceived()),
        'tribute_sent': int(player.VictoryPointsAndAttributes().VpTributeSent()),
        'total_food': int(player.VictoryPointsAndAttributes().VpTotalFood()),
        'total_wood': int(player.VictoryPointsAndAttributes().VpTotalWood()),
        'total_gold': int(player.VictoryPointsAndAttributes().VpTotalGold()),
        'total_stone': int(player.VictoryPointsAndAttributes().VpTotalStone()),
        'value_spent_objects': int(player.VictoryPointsAndAttributes().AttrValueSpentObjects()),
        'value_spent_research': int(player.VictoryPointsAndAttributes().AttrValueSpentResearch()),
        'value_lost_units': int(player.VictoryPointsAndAttributes().AttrValueLostUnits()),
        'value_lost_buildings': int(player.VictoryPointsAndAttributes().AttrValueLostBuildings()),
        'value_current_units': int(player.VictoryPointsAndAttributes().AttrValueCurrentUnits()),
        'value_current_buildings': int(player.VictoryPointsAndAttributes().AttrValueCurrentBuildings()),
        'value_objects_destroyed': int(player.VictoryPointsAndAttributes().AttrValueObjectsDestroyed()),
        'kills': int(player.VictoryPointsAndAttributes().AttrKills()),
        'deaths': int(player.VictoryPointsAndAttributes().AttrDeaths()),
        'razes': int(player.VictoryPointsAndAttributes().AttrRazes())
    }


def update_research(player_number, tech, research):
    """Update research structure."""
    if tech.State() == TECH_STATE_RESEARCHING and tech.IdIndex() not in research[player_number]:
        research[player_number][tech.IdIndex()] = {
            'started': tech.LastStateChange(),
            'finished': None
        }
    elif tech.State() == TECH_STATE_DONE and tech.IdIndex() in research[player_number]:
        research[player_number][tech.IdIndex()]['finished'] = tech.LastStateChange()
    elif tech.State() == TECH_STATE_AVAILABLE and tech.IdIndex() in research[player_number]:
        del research[player_number][tech.IdIndex()]


def update_market(tick, coefficients, market):
    """Update market coefficients structure."""
    last = None
    change = True
    food = coefficients.Food()
    wood = coefficients.Wood()
    stone = coefficients.Stone()
    if market:
        last = market[-1]
        change = has_diff(last, food=food, wood=wood, stone=stone)
    if change:
        market.append({
            'timestamp': tick,
            'food': food,
            'wood': wood,
            'stone': stone
        })


def add_objects(obj, objects):
    """Add objects."""
    player_number = obj.OwnerId() if obj.OwnerId() > 0 else None
    if obj.MasterObjectClass() not in [CLASS_UNIT, CLASS_BUILDING]:
        return
    if obj.Id() not in objects:
        objects[obj.Id()] = {
            'initial_player_number': player_number,
            'initial_object_id': obj.MasterObjectId(),
            'initial_class_id': obj.MasterObjectClass(),
            'created': obj.CreatedTime(),
            'destroyed': None,
            'destroyed_by_instance_id': None,
            'destroyed_building_percent': None,
            'deleted': False,
            'created_x': obj.Position().X(),
            'created_y': obj.Position().Y(),
            'destroyed_x': None,
            'destroyed_y': None,
            'building_started': None,
            'building_completed': None,
            'total_idle_time': None
        }


def update_objects(tick, obj, objects, state, last):
    """Update object/state structures."""
    player_number = obj.OwnerId() if obj.OwnerId() > 0 else None
    if obj.MasterObjectClass() not in [CLASS_UNIT, CLASS_BUILDING]:
        return
    if obj.KilledTime() > 0 and obj.Id() in objects:
        data = {
            'destroyed': obj.KilledTime(),
            'deleted': obj.DeletedByOwner()
        }
        if obj.MasterObjectClass() == CLASS_UNIT:
            data.update({
                'destroyed_x': obj.Position().X(),
                'destroyed_y': obj.Position().Y()
            })
        elif obj.MasterObjectClass() == CLASS_BUILDING:
            data['destroyed_building_percent'] = obj.BuildingPercentComplete()
        if obj.KilledByUnitId() in objects:
            data.update({
                'destroyed_by_instance_id': obj.KilledByUnitId()
            })
        objects[obj.Id()].update(data)

    if obj.MasterObjectClass() == CLASS_BUILDING and obj.Id() in objects:
        if obj.BuildingStartTime() > 0 and objects[obj.Id()]['building_started'] is None:
            objects[obj.Id()]['building_started'] = obj.BuildingStartTime()
        if obj.BuildingCompleteTime() > 0 and objects[obj.Id()]['building_completed'] is None:
            objects[obj.Id()]['building_completed'] = obj.BuildingCompleteTime()

    if obj.Id() in objects:
        objects[obj.Id()]['total_idle_time'] = int(obj.CumulativeIdleTime())

    researching_technology_id = obj.CurrentlyResearchingTechId() if obj.CurrentlyResearchingTechId() > 0 else None
    change = (
        obj.Id() not in last or
        has_diff(
            last[obj.Id()],
            player_number=player_number,
            object_id=obj.MasterObjectId(),
            researching_technology_id=researching_technology_id
        )
    )
    snapshot = {
        'timestamp': tick,
        'instance_id': obj.Id(),
        'player_number': player_number,
        'object_id': obj.MasterObjectId(),
        'class_id': obj.MasterObjectClass(),
        'researching_technology_id': researching_technology_id
    }
    if change:
        state.append(snapshot)
    last[obj.Id()] = snapshot


def enrich_actions(actions, objects, states):
    """Attach player to anonymous actions."""
    owners = defaultdict(list)
    for state in states:
        owners[state['instance_id']].append((state['timestamp'], state['player_number']))
    last = None
    for (tick, action_type, payload) in actions:
        if 'player_id' not in payload and 'object_ids' in payload:
            for object_id in payload['object_ids']:
                if object_id not in owners:
                    continue
                for (timestamp, player_number) in owners[object_id]:
                    if timestamp > tick:
                        payload['player_id'] = player_number
                        break
                if object_id not in objects:
                    continue
                if 'player_id' not in payload:
                    payload['player_id'] = objects[object_id]['initial_player_number']
                    break
        action = (tick, action_type, payload)
        if action != last and 'player_id' in payload and payload['player_id'] is not None:
            yield action
            last = action


def transform_seed_objects(objects):
    """Map seed objects to state format."""
    return {obj['instance_id']: {
        'initial_player_number': obj['player_number'],
        'initial_object_id': obj['object_id'],
        'initial_class_id': obj['class_id'],
        'created': 0,
        'created_x': obj['x'],
        'created_y':obj['y'],
        'destroyed': None,
        'destroyed_by_instance_id': None,
        'destroyed_building_percent': None,
        'deleted': False,
        'destroyed_x': None,
        'destroyed_y': None,
        'building_started': None,
        'building_completed': None,
        'total_idle_time': None
    } for obj in objects}


async def get_extracted_data(start_time, duration, playback, handle, interval, seed_objects):
    """Get extracted data."""
    timeseries = []
    research = defaultdict(dict)
    market = []
    objects = transform_seed_objects(seed_objects)
    state = []
    last = {}
    actions = []
    tribute = []
    transactions = []
    formations = []
    version = None
    start = time.time()
    client = await Client.create(playback, handle.name, start_time, duration, interval)
    async for tick, source, message in client.sync():
        if source == Source.MEMORY:
            if not version:
                version = message.StateReaderVersion().decode('ascii')

            update_market(tick, message.MarketCoefficients(), market)

            # Add any new objects before updating to ensure fks are present for updates
            for i in range(0, message.ObjectsLength()):
                add_objects(message.Objects(i), objects)

            for i in range(0, message.ObjectsLength()):
                update_objects(tick, message.Objects(i), objects, state, last)

            for i in range(0, message.PlayersLength()):
                player = message.Players(i)
                timeseries.append(build_timeseries_record(tick, player))
                for j in range(0, player.TechsLength()):
                    update_research(player.Id(), player.Techs(j), research)

        elif source == Source.MGZ:
            if message[0] == fast.Operation.ACTION:
                actions.append((tick, *message[1]))
                if message[1][0] == fast.Action.TRIBUTE:
                    tribute.append((tick, message[1][1]))
                elif message[1][0] in [fast.Action.BUY, fast.Action.SELL]:
                    transactions.append((tick, *message[1]))
                elif message[1][0] == fast.Action.FORMATION:
                    formations.append((tick, *message[1]))

    handle.close()

    return {
        'version': version,
        'runtime': timedelta(seconds=int(time.time() - start)),
        'timeseries': timeseries,
        'research': flatten_research(research),
        'market': market,
        'objects': [dict(obj, instance_id=i) for i, obj in objects.items()],
        'state': state,
        'actions': list(enrich_actions(actions, objects, state)),
        'tribute': tribute,
        'transactions': transactions
    }



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: AOC

class ObjectState(object):
    Ready = 0
    Founded = 1
    Alive = 2
    Dying = 3
    Dead = 4
    Rotting = 5
    Unknown = 6
    Dopple = 7
    Reusable = 8

